#region GNU_GPL
// Copyright (c) 2003 The Ohio State University
//
// Anish Arora <anish@cis.ohio-state.edu>
// Jason Hallstrom <hallstro@cis.ohio-state.edu>
// Bill Leal <leal@cis.ohio-state.edu>
//
// This file is part of DRSS.
//
// DRSS is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// DRSS is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with DRSS; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#endregion


using System;
using System.IO;
using System.Threading;
using System.Net.Sockets;
using System.Runtime.Serialization.Formatters.Binary;
using OSU.DRSS.Util;


namespace OSU.DRSS.Platform.Messaging.Transport
{
	
	public class TCPReceiver
	{
		#region Fields
		//-----------------------------------------------------------------
		// Fields
		//-----------------------------------------------------------------		
		// All Message objects received will be delivered to this target
		private MessageReceiver mrReceiverTarget;

		// The port number on which to listen for connection requests
		private int intPortNumber;

		// The listener used to accept socket connections
		private TcpListener tcpListener;

		// The number of active socket connections
		private int intConnectionCount;

		// A dummy object used to ensure mutually exclusive access to intConnectionCount
		private object mtx_intConnectionCount;

		// The maximum time (in milliseconds) that a receive can be in progress without being
		// canceled
		private int intReceiveTimeout;

		// The maximum time (in milliseconds) that a socket can be inactive without
		// being closed by the TCPReceiver
		private int intInactivityTimeout;
		//-----------------------------------------------------------------		
		#endregion


		#region Constructors
		//-----------------------------------------------------------------		
		// Constructors
		//-----------------------------------------------------------------	
		public TCPReceiver(MessageReceiver mrReceiverTarget, int intPortNumber, int intReceiveTimeout, int intInactivityTimeout)
		{ 
			// Note that mrReceiverTarget must be thread-safe
			this.mrReceiverTarget = mrReceiverTarget;
			this.intPortNumber = intPortNumber;
			this.tcpListener = null;
			this.intConnectionCount = 0;
			this.mtx_intConnectionCount = new object();
			this.intReceiveTimeout = intReceiveTimeout;
			this.intInactivityTimeout = intInactivityTimeout;

			// Bind the listener socket, and begin accepting connections
			tcpListener = new TcpListener(intPortNumber);
			tcpListener.Start();
			
			// Run AcceptThreadBody() in a separate thread
			Thread thdAcceptThread = new Thread(new ThreadStart(AcceptThreadBody));
			thdAcceptThread.IsBackground = true;
			thdAcceptThread.Start();
		}
		//-----------------------------------------------------------------	
		#endregion


		#region Properties
		//-----------------------------------------------------------------		
		// Properties
		//-----------------------------------------------------------------	
		public int ConnectionCount
		{
			get
			{	
				lock(mtx_intConnectionCount)
				{	return(intConnectionCount);   }
			}
		}
		//-----------------------------------------------------------------		
		#endregion


		#region Methods
		//-----------------------------------------------------------------
		// Methods
		//-----------------------------------------------------------------		
		private void AcceptThreadBody()
		{
			// Continue accepting connections until the thread is killed
			while(true)
			{
				try
				{
					// Accept a client connection, and set the receive timeout
					TcpClient tcpClient = tcpListener.AcceptTcpClient();
					tcpClient.ReceiveTimeout = intReceiveTimeout;
					
					lock(mtx_intConnectionCount)
					{	intConnectionCount++;   }
					
					// Run ReceiveThreadBody() in a separate thread
					Thread thdReceiveThread = 
						ThreadWithArgumentsFactory.CreateThread(new ThreadWithArgumentsStart(ReceiveThreadBody), tcpClient);
					thdReceiveThread.IsBackground = true;
					thdReceiveThread.Start();
				}
				catch(SocketException se)
				{
					// Forward the exception
					// (Ignore the exception raised when the tcpListener is closed)
					if((se.ErrorCode != 10004) && (ConnectExceptionEvent != null))
					{	ConnectExceptionEvent(this, new ExceptionEventArgs(se));   }
				}
				catch(Exception exc)
				{   
					// Forward the exception
					if(ConnectExceptionEvent != null)
					{	ConnectExceptionEvent(this, new ExceptionEventArgs(exc));   }
				}
			}  
		}

		private void ReceiveThreadBody(object obj_tcpClient)
		{
			TcpClient tcpClient = (TcpClient) obj_tcpClient;
			long lonTimeLastActive = DateTime.Now.Ticks / 10000;
	
			// Continue reading Message objects until the thread is killed, or the
			// socket times out due to inactivity
			bool bolTimeout = false;
			while(!bolTimeout)
			{
				try
				{
					// Deserialize the Message object
					BinaryFormatter bfFormatter = new BinaryFormatter();
					Message msg = (Message) bfFormatter.Deserialize(tcpClient.GetStream());

					// Deliver the message to the target
					mrReceiverTarget.ReceiveMessage(msg);
					lonTimeLastActive = DateTime.Now.Ticks / 10000;
				}
				catch(IOException ioe)
				{
					// Forward the exception
					// (Do not forward the exception raised when the tcpClient times out)
					if((ioe.InnerException == null) || 
						(!ioe.InnerException.GetType().Equals(typeof(SocketException))) ||
						(((SocketException)ioe.InnerException).ErrorCode != 10060))
					{
						if(ReceiveExceptionEvent != null)
						{	ReceiveExceptionEvent(this, new ExceptionEventArgs(ioe));   }
					}
				}
				catch(Exception exc)
				{
					// Forward the exception
					if(ReceiveExceptionEvent != null)
					{	ReceiveExceptionEvent(this, new ExceptionEventArgs(exc));   }
				}

				// Determine whether the socket has been inactive for too long
				long lonTimeNow = DateTime.Now.Ticks / 10000;
				if((lonTimeNow - lonTimeLastActive) > intInactivityTimeout)
				{	bolTimeout = true;   }
			}

			// Terminate the connection
			try
			{   
				lock(mtx_intConnectionCount)
				{	intConnectionCount--;   }
					
				tcpClient.Close();   
			}
			catch(Exception exc)
			{	
				// Forward the exception
				if(CloseExceptionEvent != null)
				{	CloseExceptionEvent(this, new ExceptionEventArgs(exc));   }
			}
		}
		//-----------------------------------------------------------------		
		#endregion


		#region Events
		//-----------------------------------------------------------------
		// Events
		//-----------------------------------------------------------------
		// This event is generated whenever an exception is raised while establishing a connection
		public event ExceptionEventDelegate ConnectExceptionEvent;

		// This event is generated whenever an exception is raised while receiving on a connection
		public event ExceptionEventDelegate ReceiveExceptionEvent;

		// This exception is generated whenever an exception is raised while closing a connection
		public event ExceptionEventDelegate CloseExceptionEvent;
		//-----------------------------------------------------------------		
		#endregion
	}
}
