#region GNU_GPL
// Copyright (c) 2003 The Ohio State University
//
// Anish Arora <anish@cis.ohio-state.edu>
// Jason Hallstrom <hallstro@cis.ohio-state.edu>
// Bill Leal <leal@cis.ohio-state.edu>
//
// This file is part of DRSS.
//
// DRSS is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// DRSS is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with DRSS; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#endregion


using System;
using System.Net;
using System.Threading;
using System.Net.Sockets;
using System.Collections;
using System.Runtime.Serialization.Formatters.Binary;
using OSU.DRSS.Util;


namespace OSU.DRSS.Platform.Messaging.Transport
{
	
	public class TCPSender : MessageSender
	{
		private class TCPClientSelectionKey
		{
			#region Fields
			//-----------------------------------------------------------------
			// Fields
			//-----------------------------------------------------------------
			// The thread ID portion of the selection key
			private int intThreadID;
		
			// The IP end-point portion of the selection key
			private IPEndPoint ipeIPEndPoint;
			//-----------------------------------------------------------------
			#endregion


			#region Constructors
			//-----------------------------------------------------------------		
			// Constructors
			//-----------------------------------------------------------------	
			public TCPClientSelectionKey(int intThreadID, IPEndPoint ipeIPEndPoint)
			{
				this.intThreadID = intThreadID;
				this.ipeIPEndPoint = ipeIPEndPoint;
			}
			//-----------------------------------------------------------------	
			#endregion


			#region Methods
			//-----------------------------------------------------------------
			// Methods
			//-----------------------------------------------------------------		
			public override bool Equals(object objSelectionKey)
			{
				// The keys are equal if both key parts match
				TCPClientSelectionKey tcpOtherKey = (TCPClientSelectionKey) objSelectionKey;
				return((this.intThreadID == tcpOtherKey.intThreadID) &&
					   (this.ipeIPEndPoint.Equals(tcpOtherKey.ipeIPEndPoint)));
			}

			public override int GetHashCode()
			{	return(intThreadID.GetHashCode() ^ ipeIPEndPoint.GetHashCode());   }
			//-----------------------------------------------------------------		
			#endregion
		}


		private class TCPClientUsedPair
		{
			#region Fields
			//-----------------------------------------------------------------
			// Fields
			//-----------------------------------------------------------------
			// The TCP client
			private TcpClient tcpClient;

			// A flag indicating whether data was sent on the TCP client since the
			// last run of the socket cleanup thread
			private bool bolUsed;
			//-----------------------------------------------------------------
			#endregion


			#region Constructors
			//-----------------------------------------------------------------		
			// Constructors
			//-----------------------------------------------------------------	
			public TCPClientUsedPair(TcpClient tcpClient)
			{
				this.tcpClient = tcpClient;
				this.bolUsed = true;
			}
			//-----------------------------------------------------------------	
			#endregion


			#region Properties
			//-----------------------------------------------------------------		
			// Properties
			//-----------------------------------------------------------------	
			public TcpClient TCPClient
			{
				get
				{	return(tcpClient);   }
			}

			public bool Used
			{
				get
				{	return(bolUsed);   }

				set
				{	this.bolUsed = value;   }
			}
			//-----------------------------------------------------------------	
			#endregion
		}

	
		#region Fields
		//-----------------------------------------------------------------
		// Fields
		//-----------------------------------------------------------------
		// The selection table which maps TCPClientSelectionKey objects to
		// TCPClientUsedPair objects
		private Hashtable htSelectionTable;
		
		// The number of active socket connections
		private int intConnectionCount;

		// A dummy object used to ensure mutually exclusive access to intConnectionCount
		private object mtx_intConnectionCount;

		// The maximum time (in milliseconds) that a send can be in progress without being
		// canceled
		private int intSendTimeout;

		// The maximum time (in milliseconds) that a socket can be inactive without
		// being closed by the TCPSender
		private int intInactivityTimeout;
		//-----------------------------------------------------------------
		#endregion


		#region Constructors
		//-----------------------------------------------------------------		
		// Constructors
		//-----------------------------------------------------------------	
		public TCPSender(int intSendTimeout, int intInactivityTimeout)
		{
			this.htSelectionTable = new Hashtable();
			this.intConnectionCount = 0;
			this.mtx_intConnectionCount = new object();
			this.intSendTimeout = intSendTimeout;
			this.intInactivityTimeout = intInactivityTimeout;

			// Startup the socket cleanup thread
			Thread thdSocketCleanupThread = new Thread(new ThreadStart(SocketCleanupThreadBody));
			thdSocketCleanupThread.IsBackground = true;
			thdSocketCleanupThread.Start();
		}
		//-----------------------------------------------------------------	
		#endregion


		#region Properties
		//-----------------------------------------------------------------		
		// Properties
		//-----------------------------------------------------------------	
		public int ConnectionCount
		{
			get
			{	
				lock(mtx_intConnectionCount)
				{	return(intConnectionCount);   }
			}
		}
		//-----------------------------------------------------------------		
		#endregion


		#region Methods
		//-----------------------------------------------------------------
		// Methods
		//-----------------------------------------------------------------
		private void SocketCleanupThreadBody()
		{
			while(true)
			{
				// Make a pass through the selection table looking for inactive sockets
				// every intInactivityTimeout milliseconds
				Thread.Sleep(intInactivityTimeout);

				lock(htSelectionTable)
				{
					Hashtable htSelectionTableCopy = (Hashtable) htSelectionTable.Clone();
					foreach(DictionaryEntry deEntry in htSelectionTableCopy)
					{
						TCPClientUsedPair tcpPair = (TCPClientUsedPair) deEntry.Value;
						lock(tcpPair)
						{
							if(!tcpPair.Used)
							{
								// The socket has been inactive too long.  Remove it from the
								// selection table and close it
								try
								{
									htSelectionTable.Remove(deEntry.Key);
									lock(mtx_intConnectionCount)
									{	intConnectionCount--;   }
									tcpPair.TCPClient.Close();
								}
								catch(Exception exc)
								{
									// Forward the exception
									if(CloseExceptionEvent != null)
									{	CloseExceptionEvent(this, new ExceptionEventArgs(exc));   }
								}
							}
							else
							{	tcpPair.Used = false;   }
						}
					}
				}
			}
		}

		public void SendMessage(int intThreadID, Message msg)
		{
			// Note that for a given destination address, one socket connection will be maintained per
			// thread ID.
			TCPClientUsedPair tcpPair = null;
			TCPClientSelectionKey tcpSelectionKey = new TCPClientSelectionKey(intThreadID, msg.DestinationAddress.PlatformIPEndPoint);
			
			//---
			// Attempt to select the connection to use for message transmission.  If no connection is available,
			// create one, and register it with the selection table.
			lock(htSelectionTable)
			{
				if(htSelectionTable.Contains(tcpSelectionKey))
				{	
					tcpPair = (TCPClientUsedPair) htSelectionTable[tcpSelectionKey];   
				}
				else
				{ 
					// Open a new TCP connection
					TcpClient tcpClient = new TcpClient(msg.DestinationAddress.PlatformIPEndPoint.Address.ToString(),
														msg.DestinationAddress.PlatformIPEndPoint.Port);
						
					// Register the connection with the selection table
					tcpPair = new TCPClientUsedPair(tcpClient);
					htSelectionTable.Add(tcpSelectionKey, tcpPair);
					lock(mtx_intConnectionCount)
					{	intConnectionCount++;   }
				}

				// Ensure mutually exclusive access to tcpPair
				Monitor.Enter(tcpPair);
			}
			//---
			
			//---
			// Attempt to send the message on the connection
			try
			{
				BinaryFormatter bfFormatter = new BinaryFormatter();
				bfFormatter.Serialize(tcpPair.TCPClient.GetStream(), msg);
				tcpPair.Used = true;
			}
			finally
			{	Monitor.Exit(tcpPair);   }
			//---
		}
		//-----------------------------------------------------------------		
		#endregion


		#region Events
		//-----------------------------------------------------------------
		// Events
		//-----------------------------------------------------------------		
		// This exception is generated whenever an exception is raised while closing a connection
		public event ExceptionEventDelegate CloseExceptionEvent;
		//-----------------------------------------------------------------		
		#endregion
	}

}
