#region GNU_GPL
// Copyright (c) 2003 The Ohio State University
//
// Anish Arora <anish@cis.ohio-state.edu>
// Jason Hallstrom <hallstro@cis.ohio-state.edu>
// Bill Leal <leal@cis.ohio-state.edu>
//
// This file is part of DRSS.
//
// DRSS is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// DRSS is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with DRSS; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#endregion


using System;
using System.Net;
using OSU.DRSS.Util;


namespace OSU.DRSS.Platform.Messaging.Transport
{
	
	public class TransportLayer : MessageSender
	{
		#region Fields
		//-----------------------------------------------------------------
		// Fields
		//-----------------------------------------------------------------
		// The IP end-point of the transport layer (receiver)
		private IPEndPoint ipeLocalEndPoint;

		// The router used to deliver local messages
		private LocalDeliveryRouter ldrLocalRouter;

		// The object responsible for sending messages to remote transport layers
		private TCPSender tcpSender;
		
		// The object responsible for receiving messages from remote transport layers
		private TCPReceiver tcpReceiver;
		//-----------------------------------------------------------------
		#endregion


		#region Constructors
		//-----------------------------------------------------------------		
		// Constructors
		//-----------------------------------------------------------------	
		public TransportLayer(int intPortNumber, int intSocketActionTimeout, int intSocketInactivityTimeout)
		{
			// HACK (?)
			IPHostEntry iphHostEntry = Dns.Resolve(Dns.GetHostName());
			this.ipeLocalEndPoint = new IPEndPoint(iphHostEntry.AddressList[0], intPortNumber);
			this.ldrLocalRouter = new LocalDeliveryRouter();
			this.tcpSender = new TCPSender(intSocketActionTimeout, intSocketInactivityTimeout/2);
			this.tcpReceiver = new TCPReceiver(ldrLocalRouter, intPortNumber, intSocketActionTimeout, intSocketInactivityTimeout);

			// Setup the event forwarding from tcpSender and tcpReceiver
			tcpReceiver.ConnectExceptionEvent += new ExceptionEventDelegate(ConnectExceptionEventForwarder);
			tcpReceiver.ReceiveExceptionEvent += new ExceptionEventDelegate(ReceiveExceptionEventForwarder);
			tcpSender.CloseExceptionEvent += new ExceptionEventDelegate(CloseExceptionEventForwarder);
			tcpReceiver.CloseExceptionEvent += new ExceptionEventDelegate(CloseExceptionEventForwarder);
		}
		//-----------------------------------------------------------------	
		#endregion


		#region Properties
		//-----------------------------------------------------------------		
		// Properties
		//-----------------------------------------------------------------	
		public IPEndPoint LocalEndPoint
		{
			get
			{	return(ipeLocalEndPoint);   }
		}
		
		public int SenderConnectionCount
		{
			get
			{	return(tcpSender.ConnectionCount);   }
		}

		public int ReceiverConnectionCount
		{
			get
			{	return(tcpReceiver.ConnectionCount);   }
		}
		//-----------------------------------------------------------------		
		#endregion


		#region Methods
		//-----------------------------------------------------------------
		// Methods
		//-----------------------------------------------------------------
		public void SendMessage(int intThreadID, Message msg)
		{
			// If the message to be sent is destined for a local object, deliver
			// the message directly.  Otherwise, use the TCPSender to deliver the
			// message to the remote platform
			if(ipeLocalEndPoint.Equals(msg.DestinationAddress.PlatformIPEndPoint))
			{
				// The message is destined for a local object -- it has therefore been
				// 'sent' to the platform correctly.  If an exception occurs, treat
				// it as a receive-type exception
				try
				{	ldrLocalRouter.ReceiveMessage(msg);   }
				catch(Exception exc)
				{	ReceiveExceptionEvent(this, new ExceptionEventArgs(exc));   }
			}
			else
			{	tcpSender.SendMessage(intThreadID, msg);   }
		}

		public void RegisterMessageReceiver(Guid guiObjectID, MessageReceiver mrReceiver)
		{	ldrLocalRouter.RegisterMessageReceiver(guiObjectID, mrReceiver);   }

		public void UnregisterMessageReceiver(Guid guiObjectID)
		{	ldrLocalRouter.UnregisterMessageReceiver(guiObjectID);   }
		//-----------------------------------------------------------------
		#endregion


		#region Events
		//-----------------------------------------------------------------
		// Events
		//-----------------------------------------------------------------
		// This event is generated whenever an exception is raised while establishing a
		// receive-type connection to a remote transport layer
		public event ExceptionEventDelegate ConnectExceptionEvent;

		// This event is generated whenever an exception is raised while receiving a message
		// from the local (or a remote) transport layer
		public event ExceptionEventDelegate ReceiveExceptionEvent;

		// This event is generated whenever an exception is raised while closing a connection
		// to a remote transport layer
		public event ExceptionEventDelegate CloseExceptionEvent;
		//-----------------------------------------------------------------		
		#endregion


		#region Event Forwarders
		//-----------------------------------------------------------------		
		//Event Forwarders
		//-----------------------------------------------------------------		
		private void ConnectExceptionEventForwarder(object objSource, ExceptionEventArgs eeaArgs)
		{
			if(ConnectExceptionEvent != null)
			{	ConnectExceptionEvent(this, eeaArgs);   }
		}

		private void ReceiveExceptionEventForwarder(object objSource, ExceptionEventArgs eeaArgs)
		{
			if(ReceiveExceptionEvent != null)
			{	ReceiveExceptionEvent(this, eeaArgs);   }
		}

		private void CloseExceptionEventForwarder(object objSource, ExceptionEventArgs eeaArgs)
		{
			if(CloseExceptionEvent != null)
			{	CloseExceptionEvent(this, eeaArgs);   }
		}
		//-----------------------------------------------------------------		
		#endregion
	}

}
