#region GNU_GPL
// Copyright (c) 2003 The Ohio State University
//
// Anish Arora <anish@cis.ohio-state.edu>
// Jason Hallstrom <hallstro@cis.ohio-state.edu>
// Bill Leal <leal@cis.ohio-state.edu>
//
// This file is part of DRSS.
//
// DRSS is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// DRSS is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with DRSS; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#endregion


using System;
using System.Net;
using System.Threading;
using System.Reflection;
using System.Collections;
using System.Runtime.Remoting.Messaging;
using System.Runtime.Remoting.Proxies;
using OSU.DRSS.Platform.Messaging;
using OSU.DRSS.ManagedObjects;
using OSU.DRSS.Util;


namespace OSU.DRSS.Platform.MethodInvocation
{

	public delegate void UnregisterObjectStubDelegate(Guid guiObjectID);


	public class ObjectStub : RealProxy, ManagedObject, MessageReceiver
	{
		#region Fields
		//-----------------------------------------------------------------
		// Fields
		//-----------------------------------------------------------------		
		// The address of the object stub
		private PlatformHostedObjectAddress phoLocalAddress;

		// The address of the invocation target (skeleton)
		private PlatformHostedObjectAddress phoTargetAddress;
		
		// The object used to send invocation requests to the invocation target (skeleton)
		private MessageSender msSender;

		// The time (in milliseconds) to wait for a response from the invocation target
		private int intResponseTimeout;

		// The delegate used to unregister the stub when it is finalized
		private UnregisterObjectStubDelegate dlgUnregisterStub;

		// A structure containing the object's unique ID, name, description, and specification
		private ManagedObjectInformation moiObjectInformation;		

		// The pending results table mapping GUID objects, representing invocation IDs, to objects of 
		// type MethodResultInformation[], representing invocation results.  The table is used to wait 
		// on the results of (possibly concurrent) method calls.  Note that objects of type Hashtable are
		// R/W synchronized
		private Hashtable htPendingResults;
		//-----------------------------------------------------------------		
		#endregion


		#region Constructors
		//-----------------------------------------------------------------		
		// Constructors
		//-----------------------------------------------------------------		
		public ObjectStub(Type typTargetType, IPEndPoint ipeLocalPlatformIPEndPoint,
						  PlatformHostedObjectAddress phoTargetAddress,
						  MessageSender msSender, int intResponseTimeout, 
						  UnregisterObjectStubDelegate dlgUnregisterStub, 
			              string strName, string strDescription, object objSpecification)
			: base(typTargetType)
		{
			Guid guiObjectID = Guid.NewGuid();
			this.phoLocalAddress = new PlatformHostedObjectAddress(ipeLocalPlatformIPEndPoint, guiObjectID);
			this.phoTargetAddress = phoTargetAddress;
			this.msSender = msSender;
			this.intResponseTimeout = intResponseTimeout;
			this.dlgUnregisterStub = dlgUnregisterStub;
			this.moiObjectInformation 
				= new ManagedObjectInformation(guiObjectID, strName, strDescription, objSpecification);
			this.htPendingResults = new Hashtable();
		}
		//-----------------------------------------------------------------		
		#endregion


		#region Properties
		//-----------------------------------------------------------------		
		// Properties
		//-----------------------------------------------------------------		
		public ManagedObjectInformation ObjectInformation
		{
			get
			{	return(moiObjectInformation);   }
		}
		//-----------------------------------------------------------------		
		#endregion


		#region Methods
		//-----------------------------------------------------------------
		// Methods
		//-----------------------------------------------------------------
		public override IMessage Invoke(IMessage imsMsg)
		{
			//---
			// Construct the invocation request message
			IMethodCallMessage imcMsg = (IMethodCallMessage) imsMsg;
			MethodCallInformation mciCall = new MethodCallInformation(imcMsg.MethodName, (Type[]) imcMsg.MethodSignature, imcMsg.Args);
			Message msgCall = new Message(phoLocalAddress, phoTargetAddress);
			msgCall.MessageContent.Push(new MessageContentLayer(mciCall, Guid.Empty));
			//---
			
			//---
			// Send the message to the invocation target, and add the appropriate 'wait-for' 
			// entry to the pending results table
			MethodResultInformation[] mriPendingResult = new MethodResultInformation[1];
			htPendingResults.Add(mciCall.InvocationID, mriPendingResult);
			msSender.SendMessage(Thread.CurrentThread.GetHashCode(), msgCall);
			//---

			//---
			// Wait for the invocation results
			lock(mriPendingResult)
			{
				if(mriPendingResult[0] == null)
				{	Monitor.Wait(mriPendingResult, intResponseTimeout);   } 
			}

			lock(htPendingResults)
			{
				// The stub is no longer waiting for invocation results, so the
				// 'wait-for' entry can be removed from the pending results table
				htPendingResults.Remove(mciCall.InvocationID);	

				if(mriPendingResult[0] == null)
				{	throw new ResponseTimeoutException(mciCall);   }
			}
			//---
			
			//---
			// Forward the invocation results to the proxy client
			if(mriPendingResult[0].IsException)
			{	return(new ReturnMessage((Exception) mriPendingResult[0].Result, imcMsg));   }
			else
			{	
				ParameterInfo[] piParameters = this.GetProxiedType().GetMethod(imcMsg.MethodName, (Type[]) imcMsg.MethodSignature).GetParameters();
				object[] objOutArguments = GenerateOutArgumentsArray(piParameters, mriPendingResult[0].Arguments);	
				return(new ReturnMessage(mriPendingResult[0].Result, objOutArguments, objOutArguments.Length, imcMsg.LogicalCallContext, imcMsg));   
			}
			//---
		}

		private object[] GenerateOutArgumentsArray(ParameterInfo[] piParameters, object[] objArguments)
		{
			// Generate the 'out' arguments array required by the ReturnMessage constructor
			ArrayList alOutArguments = new ArrayList();
			for(int intIndex = 0; intIndex < piParameters.Length; intIndex++)
			{
				if(piParameters[intIndex].IsOut)
				{	alOutArguments.Add(objArguments[intIndex]);   }
			}
			
			return(alOutArguments.ToArray());
		}

		public void ReceiveMessage(Message msg)
		{
			//---
			// Note that at this point there should be exactly one content layer on the message
			// content stack, containing the appropriate MethodResultInformation object.  The content 
			// provider of this message will always be the corresponding ObjectSkeleton instance; no 
			// additional content provider/consumer verification is required
			//---
			
			try
			{
				bool bolOutdatedResponse = false;
				MethodResultInformation mriResult = (MethodResultInformation) ((MessageContentLayer)msg.MessageContent.Pop()).LayerContent;
			
				lock(htPendingResults)
				{
					// If a call to Invoke() is waiting on these results, update the pending results table,
					// and signal the waiting client
					MethodResultInformation[] mriPendingResult = (MethodResultInformation[]) htPendingResults[mriResult.InvocationID];
					if(mriPendingResult != null)
					{	
						lock(mriPendingResult)
						{
							mriPendingResult[0] = mriResult;   
							Monitor.PulseAll(mriPendingResult);
						}
					}
					else
					{	
						bolOutdatedResponse = true;   
					}
				}

				// If an outdated response was received, raise the appropriate exception
				if(bolOutdatedResponse)
				{	throw new OutdatedResultsReceivedException(mriResult);   }
			}
			catch(Exception exc)
			{
				// Something went wrong while processing the result -- generate the appropriate event
				if(ResultProcessingExceptionEvent != null)
				{	ResultProcessingExceptionEvent(this, new ExceptionEventArgs(exc));   }
			}
		}

		public override string ToString()
		{	return(moiObjectInformation.ToString());   }

		public void Finalize()
		{	   
			// Unregister the stub
			dlgUnregisterStub(moiObjectInformation.ObjectID);   
		}
		//-----------------------------------------------------------------
		#endregion


		#region Events
		//-----------------------------------------------------------------
		// Events
		//-----------------------------------------------------------------
		// This event is generated whenever an exception is raised while processing the results of 
		// a method call
		public event ExceptionEventDelegate ResultProcessingExceptionEvent;
		//-----------------------------------------------------------------
		#endregion
	}

}
