#region GNU_GPL
// Copyright (c) 2003 The Ohio State University
//
// Anish Arora <anish@cis.ohio-state.edu>
// Jason Hallstrom <hallstro@cis.ohio-state.edu>
// Bill Leal <leal@cis.ohio-state.edu>
//
// This file is part of DRSS.
//
// DRSS is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// DRSS is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with DRSS; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#endregion


using System;
using System.Net;
using System.Threading;
using System.Collections;
using OSU.DRSS.Platform.Messaging.Interception;
using OSU.DRSS.Platform.Messaging.Transport;
using OSU.DRSS.Platform.Messaging;
using OSU.DRSS.Util;


namespace OSU.DRSS.Platform
{
	
	public class Platform : MarshalByRefObject
	{
		#region Fields
		//-----------------------------------------------------------------
		// Fields
		//-----------------------------------------------------------------		
		// The object ID of the (default-registered) singleton Platform
		public static Guid PLATFORM_ID = new Guid("9D6ACA9F-19A9-4979-96FB-B576CA977DF4");
		
		// The underlying transport layer responsible for sending and receiving messages
		private TransportLayer tlTransportLayer;
		
		// The platform's interceptor chain registry
		private PlatformInterceptorChainRegistry picRegistry;

		// The platform's stub / skeleton registry
		private PlatformStubSkeletonRegistry pssRegistry;
		
		// The queue of messages that were stalled due to content producer / consumer mismatches.
		// The platform will periodically attempt to redeliver these stalled messages.  Note that the 
		// queue instance is thread-safe
		private Queue queStalledMessages;

		// The time (in milliseconds) to wait between attempts to redeliver stalled messages
		private int intDeliveryRetryInterval;
		//-----------------------------------------------------------------		
		#endregion


		#region Constructors
		//-----------------------------------------------------------------		
		// Constructors
		//-----------------------------------------------------------------		
		public Platform(int intPortNumber)
			: this(intPortNumber, 30000, 120000, 30000, 60000)
		{}

		public Platform(int intPortNumber, int intSocketActionTimeout, int intSocketInactivityTimeout, int intDefaultResponseTimeout, int intDeliveryRetryInterval)
		{ 
			this.tlTransportLayer = new TransportLayer(intPortNumber, intSocketActionTimeout, intSocketInactivityTimeout);
			this.picRegistry = new PlatformInterceptorChainRegistry(this, tlTransportLayer);
			this.pssRegistry = new PlatformStubSkeletonRegistry(this, tlTransportLayer, picRegistry, intDefaultResponseTimeout);
			this.queStalledMessages = Queue.Synchronized(new Queue());
			this.intDeliveryRetryInterval = intDeliveryRetryInterval;

			//---
			// Startup the redelivery thread
			Thread thdRedeliverStalledMessages = new Thread(new ThreadStart(RedeliverStalledMessagesBody));
			thdRedeliverStalledMessages.IsBackground = true;
			thdRedeliverStalledMessages.Start();
			//---

			//---
			// Register the platform and each registry as active objects
			this.pssRegistry.RegisterActiveObject("Platform", "A reference to the local platform.", null, this);
			this.pssRegistry.RegisterActiveObject("IC_Registry", "A reference to the local interceptor-chain registry.", null, picRegistry);
			this.pssRegistry.RegisterActiveObject("SS_Registry", "A reference to the local stub-skeleton registry.", null, pssRegistry);
			//---

			//---
			// Setup event forwarding from the transport layer
			tlTransportLayer.ConnectExceptionEvent += new ExceptionEventDelegate(ConnectExceptionEventForwarder);
			tlTransportLayer.ReceiveExceptionEvent += new ExceptionEventDelegate(ReceiveExceptionEventForwarder);
			tlTransportLayer.CloseExceptionEvent += new ExceptionEventDelegate(CloseExceptionEventForwarder);
			//---
		}
		//-----------------------------------------------------------------		
		#endregion


		#region Properties
		//-----------------------------------------------------------------		
		// Properties
		//-----------------------------------------------------------------	
		public PlatformInterceptorChainRegistry InterceptorChainRegistry
		{
			get
			{	return(picRegistry);   }
		}

		public PlatformStubSkeletonRegistry StubSkeletonRegistry
		{
			get
			{	return(pssRegistry);   }
		}

		public IPEndPoint LocalEndPoint
		{
			get
			{	return(tlTransportLayer.LocalEndPoint);   }
		}
		
		public int SenderConnectionCount
		{
			get
			{	return(tlTransportLayer.SenderConnectionCount);   }
		}	

		public int ReceiverConnectionCount
		{
			get
			{	return(tlTransportLayer.ReceiverConnectionCount);   }
		}

		public int StalledMessageCount
		{
			get
			{	return(queStalledMessages.Count);   }
		}
		//-----------------------------------------------------------------	
		#endregion


		#region Methods
		//-----------------------------------------------------------------
		// Methods
		//-----------------------------------------------------------------		
		private void RedeliverStalledMessagesBody()
		{
			// Periodically attempt to redeliver stalled messages until the thread is killed
			while(true)
			{
				Thread.Sleep(intDeliveryRetryInterval);

				// Attempt to redeliver each message in the queue.  As each message is destined for a local
				// object, the transport layer will generate exception events if something goes wrong with
				// redelivery.  No exceptions will be raised by TransportLayer.SendMessage()
				int intCount = queStalledMessages.Count;
				while(intCount > 0)
				{
					Message msgStalledMessage = (Message) queStalledMessages.Dequeue();
					tlTransportLayer.SendMessage(Thread.CurrentThread.GetHashCode(), msgStalledMessage);
					intCount--;
				}
			}
		}
		//-----------------------------------------------------------------		
		#endregion


		#region Events
		//-----------------------------------------------------------------
		// Events
		//-----------------------------------------------------------------
		// This event is generated whenever an exception is raised while establishing a
		// receive-type connection to a remote platform
		public event ExceptionEventDelegate ConnectExceptionEvent;

		// This event is generated whenever an exception is raised while receiving a message
		// from the local (or a remote) platform
		public event ExceptionEventDelegate ReceiveExceptionEvent;

		// This event is generated whenever an exception is raised while closing a connection
		// to a remote platform
		public event ExceptionEventDelegate CloseExceptionEvent;
		//-----------------------------------------------------------------		
		#endregion


		#region Event Forwarders
		//-----------------------------------------------------------------		
		//Event Forwarders
		//-----------------------------------------------------------------		
		private void ConnectExceptionEventForwarder(object objSource, ExceptionEventArgs eeaArgs)
		{
			if(ConnectExceptionEvent != null)
			{	ConnectExceptionEvent(this, eeaArgs);   }
		}

		private void ReceiveExceptionEventForwarder(object objSource, ExceptionEventArgs eeaArgs)
		{
			if(ReceiveExceptionEvent != null)
			{	ReceiveExceptionEvent(this, eeaArgs);   }
			
			// If a message was undeliverable due to a content producer / consumer mismatch, stall
			// the message.  The platform will periodically attempt to redeliver these stalled messages
			if(eeaArgs.Exception.GetType().Equals(typeof(UnrecognizedMessageContentException)))
			{	
				UnrecognizedMessageContentException umcExc = (UnrecognizedMessageContentException) eeaArgs.Exception;
				queStalledMessages.Enqueue(umcExc.UnrecognizedMessage);
			}
		}

		private void CloseExceptionEventForwarder(object objSource, ExceptionEventArgs eeaArgs)
		{
			if(CloseExceptionEvent != null)
			{	CloseExceptionEvent(this, eeaArgs);   }
		}
		//-----------------------------------------------------------------		
		#endregion
	}

}
