#region GNU_GPL
// Copyright (c) 2003 The Ohio State University
//
// Anish Arora <anish@cis.ohio-state.edu>
// Jason Hallstrom <hallstro@cis.ohio-state.edu>
// Bill Leal <leal@cis.ohio-state.edu>
//
// This file is part of DRSS.
//
// DRSS is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// DRSS is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with DRSS; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#endregion


using System;
using System.Collections;
using OSU.DRSS.Platform.Messaging.Interception;
using OSU.DRSS.Platform.Messaging.Transport;
using OSU.DRSS.ManagedObjects;


namespace OSU.DRSS.Platform
{

	public class PlatformInterceptorChainRegistry : MarshalByRefObject
	{
		#region Fields
		//-----------------------------------------------------------------
		// Fields
		//-----------------------------------------------------------------		
		// The object ID of the (default-registered) singleton PlatformInterceptorChainRegistry
		public static Guid IC_REGISTRY_ID = new Guid("E09D2182-A487-4ab3-8578-4DCE3E8EE942");

		// The local platform containing this registry
		private Platform plaLocalPlatform;
		
		// The underlying transport layer of the local platform
		private TransportLayer tlTransportLayer;
		
		// A mapping from Guid to SendInterceptorChain.  This table is used to keep track of
		// registered send interceptor chains
		private Hashtable htSIChains;

		// The default SendInterceptorChain instance
		private SendInterceptorChain sicDefaultSendChain;

		// A mapping from Guid to Hashtable, where each inner Hashtable provides a mapping from
		// GUID to ManagedObjectInformation.  (Chain ID --> (Object ID --> ManagedObjectInformation)).
		// This table is used to keep track of which objects depend on each send interceptor chain
		private Hashtable htSICDependents;

		// A mapping from Guid to ReceiveInterceptorChain.  This table is used to keep track of
		// registered receive interceptor chains
		private Hashtable htRIChains;

		// The default ReceiveInterceptorChain instance
		private ReceiveInterceptorChain ricDefaultReceiveChain;

		// A mapping from Guid to Hashtable, where each inner Hashtable provides a mapping from
		// GUID to ManagedObjectInformation.  (Chain ID --> (Object ID --> ManagedObjectInformation)).
		// This table is used to keep track of which objects depend on each receive interceptor chain
		private Hashtable htRICDependents;
		//-----------------------------------------------------------------		
		#endregion
		

		#region Constructors
		//-----------------------------------------------------------------		
		// Constructors
		//-----------------------------------------------------------------		
		public PlatformInterceptorChainRegistry(Platform plaLocalPlatform, TransportLayer tlTransportLayer)
		{
			this.plaLocalPlatform = plaLocalPlatform;
			this.tlTransportLayer = tlTransportLayer;
			this.htSIChains = new Hashtable();
			this.htSICDependents = new Hashtable();
			this.htRIChains = new Hashtable();
			this.htRICDependents = new Hashtable();

			//---
			// Setup the default interceptor chains
			this.sicDefaultSendChain = new SendInterceptorChain(tlTransportLayer, "Default", "The default send-interceptor chain.", null);
			this.htSIChains.Add(sicDefaultSendChain.ObjectInformation.ObjectID, sicDefaultSendChain);
			this.htSICDependents.Add(sicDefaultSendChain.ObjectInformation.ObjectID, new Hashtable());

			this.ricDefaultReceiveChain = new ReceiveInterceptorChain("Default", "The default receive-interceptor chain.", null);
			this.htRIChains.Add(ricDefaultReceiveChain.ObjectInformation.ObjectID, ricDefaultReceiveChain);
			this.htRICDependents.Add(ricDefaultReceiveChain.ObjectInformation.ObjectID, new Hashtable());
			//---
		}
		//-----------------------------------------------------------------		
		#endregion


		#region SI Chain Management Methods
		//-----------------------------------------------------------------		
		// SI Chain Management Methods
		//-----------------------------------------------------------------		
		public Guid RegisterSIChain(string strName, string strDescription, object objSpecification)
		{
			lock(plaLocalPlatform)
			{
				SendInterceptorChain sicChain = new SendInterceptorChain(tlTransportLayer, strName, strDescription, objSpecification);
				RegisterChain(sicChain, htSIChains, htSICDependents);
				
				//---
				// Generate the appropriate event
				if(SIChainRegisteredEvent != null)
				{	SIChainRegisteredEvent(this, new ChainRegisteredEventArgs(sicChain.ObjectInformation));   }
				//---

				return(sicChain.ObjectInformation.ObjectID);
			}
		}
		
		public void UnregisterSIChain(Guid guiChainID)
		{
			lock(plaLocalPlatform)
			{	
				UnregisterChain(guiChainID, htSIChains, htSICDependents);   
			
				//---
				// Generate the appropriate event
				if(SIChainUnregisteredEvent != null)
				{	SIChainUnregisteredEvent(this, new ChainUnregisteredEventArgs(guiChainID));   }
				//---
			}
		}

		public IEnumerator GetSIChains()
		{
			lock(plaLocalPlatform)
			{	return(MOIEnumeratorFactory.GetMOIEnumerator(htSIChains.Values));   }
		}

		public ManagedObjectInformation GetSIChain(Guid guiChainID)
		{
			lock(plaLocalPlatform)
			{	return(GetChain(guiChainID, htSIChains));   }
		}

		public IEnumerator GetMatchingSIChains(string strName)
		{
			lock(plaLocalPlatform)
			{	return(GetMatchingChains(strName, htSIChains));   }
		}

		public IEnumerator GetSIChainElements(Guid guiChainID)
		{
			lock(plaLocalPlatform)
			{	return(GetChainElements(guiChainID, htSIChains));   }
		}

		public IEnumerator GetSIChainDependents(Guid guiChainID)
		{	
			lock(plaLocalPlatform)
			{	return(GetChainDependents(guiChainID, htSICDependents));   }
		}
		//-----------------------------------------------------------------		
		#endregion


		#region RI Chain Management Methods
		//-----------------------------------------------------------------		
		// RI Chain Management Methods
		//-----------------------------------------------------------------		
		public Guid RegisterRIChain(string strName, string strDescription, object objSpecification)
		{
			lock(plaLocalPlatform)
			{
				ReceiveInterceptorChain ricChain = new ReceiveInterceptorChain(strName, strDescription, objSpecification);
				RegisterChain(ricChain, htRIChains, htRICDependents);

				//---
				// Generate the appropriate event
				if(RIChainRegisteredEvent != null)
				{	RIChainRegisteredEvent(this, new ChainRegisteredEventArgs(ricChain.ObjectInformation));   }
				//---

				return(ricChain.ObjectInformation.ObjectID);
			}
		}
		
		public void UnregisterRIChain(Guid guiChainID)
		{
			lock(plaLocalPlatform)
			{	
				UnregisterChain(guiChainID, htRIChains, htRICDependents);   
			
				//---
				// Generate the appropriate event
				if(RIChainUnregisteredEvent != null)
				{	RIChainUnregisteredEvent(this, new ChainUnregisteredEventArgs(guiChainID));   }
				//---
			}
		}

		public IEnumerator GetRIChains()
		{
			lock(plaLocalPlatform)
			{	return(MOIEnumeratorFactory.GetMOIEnumerator(htRIChains.Values));   }
		}

		public ManagedObjectInformation GetRIChain(Guid guiChainID)
		{
			lock(plaLocalPlatform)
			{	return(GetChain(guiChainID, htRIChains));   }
		}

		public IEnumerator GetMatchingRIChains(string strName)
		{
			lock(plaLocalPlatform)
			{	return(GetMatchingChains(strName, htRIChains));   }
		}

		public IEnumerator GetRIChainElements(Guid guiChainID)
		{
			lock(plaLocalPlatform)
			{	return(GetChainElements(guiChainID, htRIChains));   }
		}

		public IEnumerator GetRIChainDependents(Guid guiChainID)
		{	
			lock(plaLocalPlatform)
			{	return(GetChainDependents(guiChainID, htRICDependents));   }
		}
		//-----------------------------------------------------------------		
		#endregion


		#region (Private) Chain Management Methods
		//-----------------------------------------------------------------		
		// (Private) Chain Management Methods
		//-----------------------------------------------------------------		
		private void RegisterChain(ManagedObject moChain, Hashtable htChains, Hashtable htDependents)
		{
			htChains.Add(moChain.ObjectInformation.ObjectID, moChain);
			htDependents.Add(moChain.ObjectInformation.ObjectID, new Hashtable());
		}

		private void UnregisterChain(Guid guiChainID, Hashtable htChains, Hashtable htDependents)
		{
			// Make sure that the chain ID corresponds to a valid chain instance
			if(!htChains.ContainsKey(guiChainID))
			{	throw new InvalidObjectIDException(guiChainID);   }

			// Make sure that no objects depend on the chain
			if(((ArrayList)htDependents[guiChainID]).Count > 0)
			{	throw new DependencyViolationException();   }

			// Make sure that the chain is not a default chain
			if((htChains[guiChainID] == sicDefaultSendChain) ||
			   (htChains[guiChainID] == ricDefaultReceiveChain))
			{	throw new DependencyViolationException("Default interceptor chains cannot be unregistered.");   }

			// Unregister the chain
			htChains.Remove(guiChainID);
			htDependents.Remove(guiChainID);
		}

		private ManagedObjectInformation GetChain(Guid guiChainID, Hashtable htChains)
		{
			// Make sure that the chain ID corresponds to a valid chain instance
			if(!htChains.ContainsKey(guiChainID))
			{	throw new InvalidObjectIDException(guiChainID);   }

			return(((ManagedObject) htChains[guiChainID]).ObjectInformation);
		}

		private IEnumerator GetMatchingChains(string strName, Hashtable htChains)
		{
			// Construct an array containing the ManagedObjectInformation for each
			// interceptor chain with the specified name
			ArrayList alMOI = new ArrayList();
			foreach(DictionaryEntry deEntry in htChains)
			{
				ManagedObject mo = (ManagedObject) deEntry.Value;
				if(mo.ObjectInformation.Name.Equals(strName))
				{	alMOI.Add(mo.ObjectInformation);   }
			}
				
			return(alMOI.GetEnumerator());
		}

		private IEnumerator GetChainElements(Guid guiChainID, Hashtable htChains)
		{
			// Make sure that the chain ID corresponds to a valid chain instance
			if(!htChains.ContainsKey(guiChainID))
			{	throw new InvalidObjectIDException(guiChainID);   }

			return(((IEnumerable)htChains[guiChainID]).GetEnumerator());
		}

		private IEnumerator GetChainDependents(Guid guiChainID, Hashtable htDependents)
		{
			// Make sure that the chain ID corresponds to a valid chain instance
			if(!htDependents.ContainsKey(guiChainID))
			{	throw new InvalidObjectIDException(guiChainID);   }

			// Return an enumeration containing the ManagedObjectInformation for each 
			// dependent object
			return(((Hashtable)htDependents[guiChainID]).Values.GetEnumerator());
		}
		//-----------------------------------------------------------------		
		#endregion


		#region Send Interceptor Management Methods
		//-----------------------------------------------------------------		
		// Send Interceptor Management Methods
		//-----------------------------------------------------------------		
		public Guid RegisterSendInterceptor(int intIndex, Type typInterceptor, params object[] objArguments)
		{
			// HACK!!!
			lock(plaLocalPlatform)
			{	
				return(RegisterSendInterceptor(sicDefaultSendChain.ObjectInformation.ObjectID, intIndex,
											   typInterceptor, objArguments));
			}
		}

		public Guid RegisterSendInterceptor(Guid guiChainID, int intIndex,
											Type typInterceptor, params object[] objArguments)
		{
			lock(plaLocalPlatform)
			{	
				ManagedObject moInterceptor = CreateAndRegisterInterceptor(guiChainID, intIndex, 
																		   typInterceptor, objArguments, 
																		   typeof(SendInterceptor), htSIChains);

				//---
				// Generate the appropriate event
				if(SIRegisteredEvent != null)
				{	SIRegisteredEvent(this, new InterceptorRegisteredEventArgs(guiChainID, intIndex, moInterceptor.ObjectInformation));   }
				//---

				return(moInterceptor.ObjectInformation.ObjectID);   
			}
		}

		public void UnregisterSendInterceptor(int intIndex)
		{
			// HACK!!!
			lock(plaLocalPlatform)
			{	UnregisterSendInterceptor(sicDefaultSendChain.ObjectInformation.ObjectID, intIndex);   }
		}

		public void UnregisterSendInterceptor(Guid guiChainID, int intIndex)
		{
			lock(plaLocalPlatform)
			{	
				UnregisterInterceptor(guiChainID, intIndex, htSIChains);   
				
				//---
				// Generate the appropriate event
				if(SIUnregisteredEvent != null)
				{	SIUnregisteredEvent(this, new InterceptorUnregisteredEventArgs(guiChainID, intIndex));   }
				//---
			}
		}
		//-----------------------------------------------------------------		
		#endregion


		#region Receive Interceptor Management Methods
		//-----------------------------------------------------------------		
		// Receive Interceptor Management Methods
		//-----------------------------------------------------------------		
		public Guid RegisterReceiveInterceptor(int intIndex, Type typInterceptor, params object[] objArguments)
		{
			// HACK!!!
			lock(plaLocalPlatform)
			{	
				return(RegisterReceiveInterceptor(ricDefaultReceiveChain.ObjectInformation.ObjectID, intIndex,
												  typInterceptor, objArguments));
			}
		}

		public Guid RegisterReceiveInterceptor(Guid guiChainID, int intIndex,
											   Type typInterceptor, params object[] objArguments)
		{
			lock(plaLocalPlatform)
			{	
				ManagedObject moInterceptor = CreateAndRegisterInterceptor(guiChainID, intIndex, 
																		   typInterceptor, objArguments, 
																		   typeof(ReceiveInterceptor), htRIChains);

				//---
				// Generate the appropriate event
				if(RIRegisteredEvent != null)
				{	RIRegisteredEvent(this, new InterceptorRegisteredEventArgs(guiChainID, intIndex, moInterceptor.ObjectInformation));   }
				//---
				
				return(moInterceptor.ObjectInformation.ObjectID);
			}
		}

		public void UnregisterReceiveInterceptor(int intIndex)
		{
			// HACK!!!
			lock(plaLocalPlatform)
			{	UnregisterReceiveInterceptor(ricDefaultReceiveChain.ObjectInformation.ObjectID, intIndex);   }
		}

		public void UnregisterReceiveInterceptor(Guid guiChainID, int intIndex)
		{
			lock(plaLocalPlatform)
			{	
				UnregisterInterceptor(guiChainID, intIndex, htRIChains);   

				//---
				// Generate the appropriate event
				if(RIUnregisteredEvent != null)
				{	RIUnregisteredEvent(this, new InterceptorUnregisteredEventArgs(guiChainID, intIndex));   }
				//---
			}
		}
		//-----------------------------------------------------------------		
		#endregion


		#region (Private) Interceptor Management Methods
		//-----------------------------------------------------------------		
		// (Private) Interceptor Management Methods
		//-----------------------------------------------------------------		
		private ManagedObject CreateAndRegisterInterceptor(Guid guiChainID, int intIndex, 
														   Type typInterceptor, object[] objArguments,
														   Type typRequiredSubtype, Hashtable htChains)
		{
			// Make sure that the chain ID corresponds to a valid chain instance
			if(!htChains.ContainsKey(guiChainID))
			{	throw new InvalidObjectIDException(guiChainID);   }

			// Make sure that the interceptor type to be instantiated is a valid 
			// subtype of either SendInterceptor or ReceiveInterceptor
			if(!typRequiredSubtype.IsAssignableFrom(typInterceptor))
			{	throw new InvalidInterceptorTypeException(typInterceptor);   }

			//---
			// Instantiate the interceptor.  Try prefixing the argument list with a references to the 
			// local platform and transport layer.  If no matching constructor exists, instantiate the 
			// interceptor using the standard argument list.  If the interceptor cannot be instantiated, 
			// an appropriate exception will be raised
			ManagedObject moInterceptor = null;
			try
			{
				object[] objExtendedArguments = new Object[objArguments.Length + 2];
				objExtendedArguments[0] = plaLocalPlatform;
				objExtendedArguments[1] = tlTransportLayer;
				Array.Copy(objArguments, 0, objExtendedArguments, 2, objArguments.Length);
				moInterceptor = (ManagedObject) Activator.CreateInstance(typInterceptor, objExtendedArguments);
			}
			catch(Exception exc) 
			{	
				exc.ToString();   
				moInterceptor = (ManagedObject) Activator.CreateInstance(typInterceptor, objArguments);
			}
			//---
			
			// Insert the interceptor into the chain.  Note that an appropriate exception
			// will be generated if intIndex is out-of-range
			InterceptorChain icChain = (InterceptorChain) htChains[guiChainID];
			icChain.RegisterInterceptor(intIndex, moInterceptor);
			
			return(moInterceptor);
		}

		private void UnregisterInterceptor(Guid guiChainID, int intIndex, Hashtable htChains)
		{
			// Make sure that the chain ID corresponds to a valid chain instance
			if(!htChains.ContainsKey(guiChainID))
			{	throw new InvalidObjectIDException(guiChainID);   }

			// Unregister the interceptor
			((InterceptorChain)htChains[guiChainID]).UnregisterInterceptor(intIndex);
		}
		//-----------------------------------------------------------------
		#endregion


		#region (Internal) Chain Accessor Methods / Properties
		//-----------------------------------------------------------------		
		// (Internal) Chain Accessor Methods / Properties
		//-----------------------------------------------------------------		
		internal void RetrieveInterceptorChains(Guid guiSIChainID, Guid guiRIChainID,
											    out SendInterceptorChain sic, 
											    out ReceiveInterceptorChain ric)
		{
			lock(plaLocalPlatform)
			{
				//---
				// Make sure that the chain IDs correspond to valid chain instances
				if(!htSIChains.ContainsKey(guiSIChainID))
				{	throw new InvalidObjectIDException(guiSIChainID);   }	
				
				if(!htRIChains.ContainsKey(guiRIChainID))
				{	throw new InvalidObjectIDException(guiRIChainID);   }	
				//---

				// Retrieve the appropriate interceptor chains
				sic = (SendInterceptorChain) htSIChains[guiSIChainID];
				ric = (ReceiveInterceptorChain) htRIChains[guiRIChainID];
			}
		}

		internal SendInterceptorChain DefaultSIC
		{
			get
			{	return(sicDefaultSendChain);   }
		}

		internal ReceiveInterceptorChain DefaultRIC
		{
			get
			{	return(ricDefaultReceiveChain);   }
		}
		//-----------------------------------------------------------------		
		#endregion

		
		#region (Internal) Chain Dependents Management Methods
		//-----------------------------------------------------------------		
		// (Internal) Chain Dependents Management Methods
		//-----------------------------------------------------------------		
		internal void AddDependents(Guid guiSIChainID, Guid guiRIChainID, ManagedObjectInformation moiDependent)
		{
			// Note that guiSIChainID must correspond to a valid SI chain instance, and 
			// guiRIChainID must correspond to a valid RI chain instance
			lock(plaLocalPlatform)
			{
				((Hashtable)htSICDependents[guiSIChainID]).Add(moiDependent.ObjectID, moiDependent);
				((Hashtable)htRICDependents[guiRIChainID]).Add(moiDependent.ObjectID, moiDependent);
			}
		}

		internal void RemoveDependents(Guid guiSIChainID, Guid guiRIChainID, Guid guiDependentID)
		{
			// Note that guiSIChainID must correspond to a valid SI chain instance, 
			// guiRIChainID must correspond to a valid RI chain instance, and guiDependentID
			// must correspond to a valid dependent
			lock(plaLocalPlatform)
			{
				((Hashtable)htSICDependents[guiSIChainID]).Remove(guiDependentID);
				((Hashtable)htRICDependents[guiRIChainID]).Remove(guiDependentID);
			}
		}
		//-----------------------------------------------------------------		
		#endregion


		#region Events
		//-----------------------------------------------------------------
		// Events
		//-----------------------------------------------------------------
		// This event is generated whenever an SI chain gets registered
		public event ChainRegisteredEventDelegate SIChainRegisteredEvent;

		// This event is generated whenever an SI chain gets unregistered
		public event ChainUnregisteredEventDelegate SIChainUnregisteredEvent;

		// This event is generated whenever an RI chain gets registered
		public event ChainRegisteredEventDelegate RIChainRegisteredEvent;

		// This event is generated whenever an RI chain gets unregistered
		public event ChainUnregisteredEventDelegate RIChainUnregisteredEvent;

		// This event is generated whenever an SI interceptor is registered
		public event InterceptorRegisteredEventDelegate SIRegisteredEvent;

		// This event is generated whenever an SI interceptor is unregistered
		public event InterceptorUnregisteredEventDelegate SIUnregisteredEvent;

		// This event is generated whenever an RI interceptor is registered
		public event InterceptorRegisteredEventDelegate RIRegisteredEvent;

		// This event is generated whenever an RI interceptor is unregistered
		public event InterceptorUnregisteredEventDelegate RIUnregisteredEvent;
		//-----------------------------------------------------------------
		#endregion
	}

}
