#region GNU_GPL
// Copyright (c) 2003 The Ohio State University
//
// Anish Arora <anish@cis.ohio-state.edu>
// Jason Hallstrom <hallstro@cis.ohio-state.edu>
// Bill Leal <leal@cis.ohio-state.edu>
//
// This file is part of DRSS.
//
// DRSS is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// DRSS is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with DRSS; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#endregion


using System;
using System.Collections;
using OSU.DRSS.Platform.Messaging.Interception;
using OSU.DRSS.Platform.Messaging.Transport;
using OSU.DRSS.Platform.MethodInvocation;
using OSU.DRSS.Platform.Messaging;
using OSU.DRSS.ManagedObjects;
using OSU.DRSS.Util;


namespace OSU.DRSS.Platform
{
	
	public class PlatformStubSkeletonRegistry : MarshalByRefObject
	{
		#region Fields
		//-----------------------------------------------------------------
		// Fields
		//-----------------------------------------------------------------		
		// The object ID of the (default-registered) singleton PlatformStubSkeletonRegistry
		public static Guid SS_REGISTRY_ID = new Guid("8A172C5C-C8DA-4e96-80BF-71CAC5536638");

		// The local platform containing this registry
		private Platform plaLocalPlatform;

		// The underlying transport layer of the local platform
		private TransportLayer tlTransportLayer;
		
		// The local platform's interceptor chain registry
		private PlatformInterceptorChainRegistry picRegistry;

		// The default time (in milliseconds) to wait for a response from a target object before throwing
		// an exception through the stub.  This value can be overriden when requesting a proxy
		private int intDefaultResponseTimeout;
		
		// A mapping from Guid to ObjectStub.  This table is used to keep track of registered
		// object stubs
		private Hashtable htObjectStubs;

		// A mapping from Guid to ArrayList, where each ArrayList contains exactly two objects.
		// The first ([0]) is an instance of SendInterceptorChain; the second ([1]) is an instance of
		// ReceiveInterceptor chain.  This table is used to keep track of which interceptor chains 
		// are used by each object stub
		private Hashtable htStubDependencies;

		// A mapping from Guid to ObjectSkeleton.  This table is used to keep track of registered
		// object skeletons
		private Hashtable htObjectSkeletons;

		// A mapping from Guid to ArrayList, where each ArrayList contains exactly two objects.
		// The first ([0]) is an instance of SendInterceptorChain; the second ([1]) is an instance of
		// ReceiveInterceptor chain.  This table is used to keep track of which interceptor chains 
		// are used by each object skeleton
		private Hashtable htSkeletonDependencies;
		//-----------------------------------------------------------------		
		#endregion
		

		#region Constructors
		//-----------------------------------------------------------------		
		// Constructors
		//-----------------------------------------------------------------		
		public PlatformStubSkeletonRegistry(Platform plaLocalPlatform, TransportLayer tlTransportLayer,
									        PlatformInterceptorChainRegistry picRegistry, int intDefaultResponseTimeout)
		{
			this.plaLocalPlatform = plaLocalPlatform;
			this.tlTransportLayer = tlTransportLayer;
			this.picRegistry = picRegistry;
			this.intDefaultResponseTimeout = intDefaultResponseTimeout;
			this.htObjectStubs = new Hashtable();
			this.htStubDependencies = new Hashtable();
			this.htObjectSkeletons = new Hashtable();
			this.htSkeletonDependencies = new Hashtable();
		}
		//-----------------------------------------------------------------		
		#endregion

		
		#region Object Stub Management Methods
		//-----------------------------------------------------------------		
		// Object Stub Management Methods
		//-----------------------------------------------------------------				
		public object GetActiveObjectProxy(PlatformHostedObjectAddress phoObjectAddress, Type typObjectType)
		{
			//---
			// This method must be invoked locally.  Attempts to invoke this method on a remote platform
			// registry will raise an exception
			//---
			
			lock(this)
			{	return(GetActiveObjectProxy(phoObjectAddress, typObjectType, intDefaultResponseTimeout));   }
		}
		
		public object GetActiveObjectProxy(PlatformHostedObjectAddress phoObjectAddress, Type typObjectType, int intResponseTimeout)		
		{
			//---
			// This method must be invoked locally.  Attempts to invoke this method on a remote platform
			// registry will raise an exception
			//---
			
			lock(plaLocalPlatform)
			{
				ObjectStub osStub = CreateAndRegisterActiveObjectProxy(phoObjectAddress, typObjectType, intResponseTimeout,
																	   picRegistry.DefaultSIC, picRegistry.DefaultRIC);
				return(osStub.GetTransparentProxy());
			}
		}

		public object GetActiveObjectProxy(PlatformHostedObjectAddress phoObjectAddress, Type typObjectType,
										   Guid guiSIChainID, Guid guiRIChainID)
		{
			//---
			// This method must be invoked locally.  Attempts to invoke this method on a remote platform
			// registry will raise an exception
			//---

			lock(this)
			{	return(GetActiveObjectProxy(phoObjectAddress, typObjectType, intDefaultResponseTimeout, guiSIChainID, guiRIChainID));   }
		}

		public object GetActiveObjectProxy(PlatformHostedObjectAddress phoObjectAddress, Type typObjectType,
			                               int intResponseTimeout, Guid guiSIChainID, Guid guiRIChainID)
		{
			//---
			// This method must be invoked locally.  Attempts to invoke this method on a remote platform
			// registry will raise an exception
			//---

			lock(plaLocalPlatform)
			{
				//---
				// Retrieve the appropriate interceptor chains.  Note that an exception will
				// be raised if either of the chain IDs are invalid
				SendInterceptorChain sic; ReceiveInterceptorChain ric;
				picRegistry.RetrieveInterceptorChains(guiSIChainID, guiRIChainID, out sic, out ric);
				//---
				
				ObjectStub osStub = CreateAndRegisterActiveObjectProxy(phoObjectAddress, typObjectType, intResponseTimeout, sic, ric);
				return(osStub.GetTransparentProxy());
			}
		}

		public IEnumerator GetProxies()
		{
			lock(plaLocalPlatform)
			{	return(MOIEnumeratorFactory.GetMOIEnumerator(htObjectStubs.Values));   }
		}

		public ManagedObjectInformation GetProxy(Guid guiProxyID)
		{
			lock(plaLocalPlatform)
			{	return(GetStubOrSkeleton(guiProxyID, htObjectStubs));   }
		}

		public IEnumerator GetMatchingProxies(string strName)
		{
			lock(plaLocalPlatform)
			{	return(GetMatchingStubsOrSkeletons(strName, htObjectStubs));   }
		}

		public ArrayList GetProxyDependencies(Guid guiProxyID)
		{
			lock(plaLocalPlatform)
			{	return GetStubOrSkeletonDependencies(guiProxyID, htStubDependencies);   }
		}
		//-----------------------------------------------------------------		
		#endregion


		#region (Private) Object Stub Management Methods
		//-----------------------------------------------------------------		
		// (Private) Object Stub Management Methods
		//-----------------------------------------------------------------		
		private ObjectStub CreateAndRegisterActiveObjectProxy(PlatformHostedObjectAddress phoTargetAddress, Type typTargetType,
															  int intResponseTimeout, SendInterceptorChain sic, ReceiveInterceptorChain ric)
		{
			// Instantiate the stub
			ObjectStub osStub = new ObjectStub(typTargetType, tlTransportLayer.LocalEndPoint, phoTargetAddress,
				                               sic, intResponseTimeout, new UnregisterObjectStubDelegate(UnregisterActiveObjectProxy),
											   "Stub_" + typTargetType, "Stub for target [" + phoTargetAddress + "]", 
											   null);
			
			// Setup stub event forwarding
			osStub.ResultProcessingExceptionEvent += new ExceptionEventDelegate(ResultProcessingExceptionEventForwarder);

			// Register the stub
			RegisterStubOrSkeleton(osStub, sic, ric, htObjectStubs, htStubDependencies);
			
			//---
			// Generate the appropriate event
			if(StubRegisteredEvent != null)
			{	
				StubRegisteredEvent(this, new StubSkeletonRegisteredEventArgs(osStub.ObjectInformation, 
																			  sic.ObjectInformation.ObjectID, 
																			  ric.ObjectInformation.ObjectID));   
			}
			//---
			
			return(osStub);
		}

		private void UnregisterActiveObjectProxy(Guid guiProxyID)
		{
			lock(plaLocalPlatform)
			{	
				UnregisterStubOrSkeleton(guiProxyID, htObjectStubs, htStubDependencies);   
			
				//---
				// Generate the appropriate event
				if(StubUnregisteredEvent != null)
				{	StubUnregisteredEvent(this, new StubSkeletonUnregisteredEventArgs(guiProxyID));   }	
				//---
			}
		}
		//-----------------------------------------------------------------		
		#endregion


		#region Object Skeleton Management Methods
		//-----------------------------------------------------------------		
		// Object Skeleton Management Methods
		//-----------------------------------------------------------------	
		public Guid RegisterActiveObject(string strName, string strDescription, object objSpecification,
										 Type typTargetType, params object[] objArguments)
		{
			lock(plaLocalPlatform)
			{	
				Guid guiObjectID = CreateAndRegisterActiveObject(strName, strDescription, objSpecification,
																 picRegistry.DefaultSIC, picRegistry.DefaultRIC,
																 typTargetType, objArguments);
				return(guiObjectID);
			}
		}

		public Guid RegisterActiveObject(string strName, string strDescription, object objSpecification, object objTarget)
		{
			lock(plaLocalPlatform)
			{
				Guid guiObjectID = RegisterActiveObject(strName, strDescription, objSpecification,
														picRegistry.DefaultSIC, picRegistry.DefaultRIC, 
														objTarget);

				return(guiObjectID);
			}
		}

		public Guid RegisterActiveObject(string strName, string strDescription, object objSpecification,
									     Guid guiSIChainID, Guid guiRIChainID, 
										 Type typTargetType, params object[] objArguments)
		{
			lock(plaLocalPlatform)
			{
				
				//---
				// Retrieve the appropriate interceptor chains.  Note that an exception will
				// be raised if either of the chain IDs are invalid
				SendInterceptorChain sic; ReceiveInterceptorChain ric;
				picRegistry.RetrieveInterceptorChains(guiSIChainID, guiRIChainID, out sic, out ric);
				//---

				Guid guiObjectID = CreateAndRegisterActiveObject(strName, strDescription, objSpecification,
																 sic, ric, typTargetType, objArguments);
				return(guiObjectID);
			}
		}

		public Guid RegisterActiveObject(string strName, string strDescription, object objSpecification, 
										 Guid guiSIChainID, Guid guiRIChainID, object objTarget)
		{
			lock(plaLocalPlatform)
			{
				//---
				// Retrieve the appropriate interceptor chains.  Note that an exception will
				// be raised if either of the chain IDs are invalid
				SendInterceptorChain sic; ReceiveInterceptorChain ric;
				picRegistry.RetrieveInterceptorChains(guiSIChainID, guiRIChainID, out sic, out ric);
				//---

				Guid guiObjectID = RegisterActiveObject(strName, strDescription, objSpecification, 
													    sic, ric, objTarget);
				return(guiObjectID);
			}
		}

		public void UnregisterActiveObject(Guid guiObjectID)
		{
			lock(plaLocalPlatform)
			{	
				UnregisterStubOrSkeleton(guiObjectID, htObjectSkeletons, htSkeletonDependencies);   
				
				//---
				// Generate the appropriate event
				if(SkeletonUnregisteredEvent != null)
				{	SkeletonUnregisteredEvent(this, new StubSkeletonUnregisteredEventArgs(guiObjectID));   }	
				//---
			}
		}

		public IEnumerator GetActiveObjects()
		{
			lock(plaLocalPlatform)
			{	return(MOIEnumeratorFactory.GetMOIEnumerator(htObjectSkeletons.Values));   }
		}

		public ManagedObjectInformation GetActiveObject(Guid guiObjectID)
		{
			lock(plaLocalPlatform)
			{	return(GetStubOrSkeleton(guiObjectID, htObjectSkeletons));   }
		}

		public IEnumerator GetMatchingActiveObjects(string strName)
		{
			lock(plaLocalPlatform)
			{	return(GetMatchingStubsOrSkeletons(strName, htObjectSkeletons));   }
		}

		public ArrayList GetActiveObjectDependencies(Guid guiObjectID)
		{
			lock(plaLocalPlatform)
			{	return GetStubOrSkeletonDependencies(guiObjectID, htSkeletonDependencies);   }
		}
		//-----------------------------------------------------------------		
		#endregion

		
		#region (Private) Object Skeleton Management Methods
		//-----------------------------------------------------------------		
		// (Private) Object Skeleton Management Methods
		//-----------------------------------------------------------------		
		private Guid CreateAndRegisterActiveObject(string strName, string strDescription, object objSpecification,
												   SendInterceptorChain sic, ReceiveInterceptorChain ric,
												   Type typTargetType, object[] objArguments)
		{
			//---
			// Instantiate the target object.  Try prefixing the argument list with a reference to the 
			// local platform.  If no matching constructor exists, instantiate the target using the 
			// standard argument list.  If the object cannot beinstantiated, an appropriate exception
			// will be raised
			object objTarget = null;
			try
			{
				object[] objExtendedArguments = new Object[objArguments.Length + 1];
				objExtendedArguments[0] = plaLocalPlatform;
				Array.Copy(objArguments, 0, objExtendedArguments, 1, objArguments.Length);
				objTarget = Activator.CreateInstance(typTargetType, objExtendedArguments);
			}
			catch(Exception exc)	
			{	
				exc.ToString();
				objTarget = Activator.CreateInstance(typTargetType, objArguments);   
			}
			//---

			return(RegisterActiveObject(strName, strDescription, objSpecification, sic, ric, objTarget));
		}

		private Guid RegisterActiveObject(string strName, string strDescription, object objSpecification,
										  SendInterceptorChain sic, ReceiveInterceptorChain ric, object objTarget)
										
		{
			// Wrap the target within an object skeleton, and setup object skeleton event forwarding
			// (Note that well-known GUIDs are used when registering the platform and each registry.)			
			Guid guiObjectID;
			if(strName.Equals("Platform"))
			{	guiObjectID = Platform.PLATFORM_ID;    }
			else
				if(strName.Equals("IC_Registry"))
			{	guiObjectID = PlatformInterceptorChainRegistry.IC_REGISTRY_ID;    } 
			else
				if(strName.Equals("SS_Registry"))
			{	guiObjectID = PlatformStubSkeletonRegistry.SS_REGISTRY_ID;    }
			else
			{	guiObjectID = Guid.NewGuid();    }

			ObjectSkeleton osTarget = new ObjectSkeleton(guiObjectID, objTarget, sic, strName, "[Skeleton] " + strDescription, objSpecification);
			osTarget.CallProcessingExceptionEvent += new ExceptionEventDelegate(CallProcessingExceptionEventForwarder);
			
			// Register the skeleton
			RegisterStubOrSkeleton(osTarget, sic, ric, htObjectSkeletons, htSkeletonDependencies);
			
			//---
			// Generate the appropriate event
			if(SkeletonRegisteredEvent != null)
			{	
				SkeletonRegisteredEvent(this, new StubSkeletonRegisteredEventArgs(osTarget.ObjectInformation, 
																				  sic.ObjectInformation.ObjectID, 
																				  ric.ObjectInformation.ObjectID));   
			}
			//---
			
			return(osTarget.ObjectInformation.ObjectID);
		}
		//-----------------------------------------------------------------		
		#endregion


		#region (Private) Object Stub / Object Skeleton Management Methods
		//-----------------------------------------------------------------		
		// (Private) Object Stub / Object Skeleton Management Methods
		//-----------------------------------------------------------------		
		private void RegisterStubOrSkeleton(ManagedObject mo, SendInterceptorChain sic, ReceiveInterceptorChain ric, 
											Hashtable htRegistry, Hashtable htDependencies)
		{
			// Add the object to the appropriate registry
			htRegistry.Add(mo.ObjectInformation.ObjectID, mo);

			// Register the object with the appropriate receive-type interceptor chain
			tlTransportLayer.RegisterMessageReceiver(mo.ObjectInformation.ObjectID, ric);
			ric.RegisterSink(mo.ObjectInformation.ObjectID, (MessageReceiver) mo);

		    // Update the chain dependents information
			picRegistry.AddDependents(sic.ObjectInformation.ObjectID, ric.ObjectInformation.ObjectID, mo.ObjectInformation);
			
			// Add the object's dependency information
			ArrayList alDependencies = new ArrayList(2);
			alDependencies.Add(sic); alDependencies.Add(ric);
			htDependencies.Add(mo.ObjectInformation.ObjectID, alDependencies);
		}

		private void UnregisterStubOrSkeleton(Guid guiObjectID, Hashtable htRegistry, Hashtable htDependencies)
		{
			// Make sure that the object ID corresponds to a valid object
			if(!htRegistry.Contains(guiObjectID))
			{	throw new InvalidObjectIDException(guiObjectID);   }

			// Remove the object from the appropriate registry
			htRegistry.Remove(guiObjectID);
				
			// Unregister the object from its receive-type interceptor chain
			tlTransportLayer.UnregisterMessageReceiver(guiObjectID);
			ReceiveInterceptorChain ric = (ReceiveInterceptorChain) ((ArrayList)htDependencies[guiObjectID])[1];				
			ric.UnregisterSink(guiObjectID);
			
			// Update the chain dependents information
			Guid guiSIChainID = ((ManagedObject) ((ArrayList)htDependencies[guiObjectID])[0]).ObjectInformation.ObjectID;
			picRegistry.RemoveDependents(guiSIChainID, ric.ObjectInformation.ObjectID, guiObjectID);
				
			// Remove the object's dependency information
			htDependencies.Remove(guiObjectID);
		}

		private IEnumerator GetMatchingStubsOrSkeletons(string strName, Hashtable htRegistry)
		{
			// Construct an array containing the ManagedObjectInformation for each object
			// with the specified name
			ArrayList alMOI = new ArrayList();
			foreach(DictionaryEntry deEntry in htRegistry)
			{
				ManagedObject mo = (ManagedObject) deEntry.Value;
				if(mo.ObjectInformation.Name.Equals(strName))
				{	alMOI.Add(mo.ObjectInformation);   }
			}
				
			return(alMOI.GetEnumerator());
		}

		private ManagedObjectInformation GetStubOrSkeleton(Guid guiObjectID, Hashtable htRegistry)
		{
			// Make sure that the object ID corresponds to a valid object
			if(!htRegistry.Contains(guiObjectID))
			{	throw new InvalidObjectIDException(guiObjectID);   }

			return(((ManagedObject)htRegistry[guiObjectID]).ObjectInformation);
		}

		private ArrayList GetStubOrSkeletonDependencies(Guid guiObjectID, Hashtable htDependencies)
		{
			// Make sure that the object ID corresponds to a valid object
			if(!htDependencies.Contains(guiObjectID))
			{	throw new InvalidObjectIDException(guiObjectID);   }
			
			// Return the ManagedObjectInformation for each chain
			ArrayList alMOI = new ArrayList(2);
			alMOI.Add(((ManagedObject)((ArrayList)htDependencies[guiObjectID])[0]).ObjectInformation);
			alMOI.Add(((ManagedObject)((ArrayList)htDependencies[guiObjectID])[1]).ObjectInformation);
			return(alMOI);
		}
		//-----------------------------------------------------------------		
		#endregion


		#region Events
		//-----------------------------------------------------------------
		// Events
		//-----------------------------------------------------------------
		// This event is generated whenever an exception is raised while an object skeleton 
		// is processing a call request
		public event ExceptionEventDelegate CallProcessingExceptionEvent;

		// This event is generated whenever an exception is raised while an object proxy
		// is processing a call response
		public event ExceptionEventDelegate ResultProcessingExceptionEvent;

		// This event is generated whenever a stub is registered
		public event StubSkeletonRegisteredEventDelegate StubRegisteredEvent;

		// This event is generated whenever a stub is unregistered
		public event StubSkeletonUnregisteredEventDelegate StubUnregisteredEvent;

		// This event is generated whenever a skeleton is registered
		public event StubSkeletonRegisteredEventDelegate SkeletonRegisteredEvent;

		// This event is generated whenever a skeleton is unregistered
		public event StubSkeletonUnregisteredEventDelegate SkeletonUnregisteredEvent;
		//-----------------------------------------------------------------
		#endregion


		#region Event Forwarders
		//-----------------------------------------------------------------		
		// Event Forwarders
		//-----------------------------------------------------------------		
		private void CallProcessingExceptionEventForwarder(object objSource, ExceptionEventArgs eeaArgs)
		{
			if(CallProcessingExceptionEvent != null)
			{	CallProcessingExceptionEvent(((ManagedObject)objSource).ObjectInformation, eeaArgs);   }
		}

		private void ResultProcessingExceptionEventForwarder(object objSource, ExceptionEventArgs eeaArgs)
		{
			if(ResultProcessingExceptionEvent != null)
			{	ResultProcessingExceptionEvent(((ManagedObject)objSource).ObjectInformation, eeaArgs);   }
		}
		//-----------------------------------------------------------------		
		#endregion
		
	}

}
